
import bpy
import math
import mathutils

def lerp(a,b,t):
    return(1.0 - t) * a + b * t



def subdivBone(armature, bone, cuts,prefix ="", sufix = ""):
    newBones = []
    for i in range(0, cuts + 1):
        boneSegment = armature.data.edit_bones.new(prefix + bone.name + sufix)
        boneSegment.head = lerp(bone.head,bone.tail,(1/(cuts +1) * i))
        boneSegment.tail = lerp(bone.head,bone.tail,(1/(cuts +1) * i+1))
        boneSegment.bbone_x = bone.bbone_x * 1.5
        boneSegment.bbone_z = bone.bbone_z * 1.5
        newBones.append(boneSegment.name)
        
    return newBones

def set_BboneHandles(armature, boneName, bbone , handle , handleType = None):
    if handle == 0:
        bbone.bbone_custom_handle_start = boneName
        if handleType != None:
            bbone.bbone_handle_type_start = handleType
    if handle ==1:
        bbone.bbone_custom_handle_end = boneName
        if handleType != None:
            bbone.bbone_handle_type_end = handleType
        
    
    

def distribute_vertWeights(armature, mesh, selection, bboneName):
    
    #editbones and posebones reference
    pBones = armature.pose.bones
    eBones = armature.data.edit_bones
    #disable bbone deform
    eBones[bboneName].use_deform = False
    
    #get bbone forward vector
    bboneVector = eBones[bboneName].y_axis
    chainRes = len(selection)

    #for each vertex, project them to the vbone vector  
    for i, v in enumerate(mesh.data.vertices):
        #get id for bbone weight
        bboneGroupIndex = mesh.vertex_groups[bboneName].index
        
        #check if current vertex  is in the bendybone group 
        for g in v.groups:
                
            if bboneGroupIndex == g.group:
                #project vertex positon to bbone
                
                bBoneIndex = pBones[bboneName].bbone_segment_index(v.co)[0]
                segmentBlend = pBones[bboneName].bbone_segment_index(v.co)[1]
                #print( v.index, bBoneIndex, bboneFac)
                bboneWeight = mesh.vertex_groups[bboneName].weight(i)
                
                #check if weight group exists to create one 
                if not mesh.vertex_groups.get(selection[bBoneIndex]):
                    mesh.vertex_groups.new(name=selection[bBoneIndex])
                   
                mesh.vertex_groups[selection[bBoneIndex]].add([i], bboneWeight * (1 - segmentBlend), 'ADD')
                
                #check if weight group exists to create one
                if not mesh.vertex_groups.get(selection[min(bBoneIndex+1,chainRes -1 ) ]):
                    mesh.vertex_groups.new(name=selection[min(bBoneIndex+1,chainRes -1 ) ])
                    
                #set wegiths for segment 1
                mesh.vertex_groups[selection[min(bBoneIndex+1,chainRes -1)]].add([i], bboneWeight * (segmentBlend), 'ADD')
          
          
def follow_bbone(armature,bone,bboneName, f):
    resolution = armature.data.bones[bboneName].bbone_segments
    bpy.ops.object.mode_set(mode = 'POSE')
    pose_bbone = armature.pose.bones[bboneName] 
    pose_bone = armature.pose.bones[bone] 
    
    c = pose_bone.constraints.new("COPY_TRANSFORMS")
    c.name = "Follow Bbone"
    c.target = armature
    c.subtarget = bboneName
    c.use_bbone_shape = True
    c.head_tail = f
    segmentTransforms = pose_bbone.bbone_segment_matrix(round(f * (resolution)), rest = True)

    bpy.ops.object.mode_set(mode = 'EDIT')
    armature.data.edit_bones[bone].matrix = armature.data.edit_bones[bboneName].matrix @ segmentTransforms
    armature.data.edit_bones[bone].length = armature.data.edit_bones[bboneName].length /  armature.data.edit_bones[bboneName].bbone_segments 
    



class set_bboneStartHandleOperator(bpy.types.Operator):
    """ sets bbone start handle """
    bl_idname = "set.bbonehandle0"
    bl_label = "set bbone start handle"

    @classmethod
    def poll(cls,context):
        if context.mode == 'EDIT_ARMATURE':  
            if len(context.selected_bones) == 2 and context.active_bone.bbone_segments != 1:
                return True
    def execute(self,context):
        set_BboneHandles(context.object,context.selected_bones[1],context.active_bone, handle = 0)
        
        return{'FINISHED'}
    
class set_bboneEndHandleOperator(bpy.types.Operator):
    """ sets bbone end handle """
    bl_idname = "set.bbonehandle1"
    bl_label = "set bbone end handle"
    
    @classmethod
    def poll(cls,context):
        if context.mode == 'EDIT_ARMATURE':  
            if len(context.selected_bones) == 2 and context.active_bone.bbone_segments != 1:
                return True

    def execute(self,context):
        if len(context.selected_bones) == 2:
            set_BboneHandles(context.object,context.selected_bones[1],context.active_bone, handle = 1)
        
        return{'FINISHED'}

class convert_bbonesOperator(bpy.types.Operator):
    """ converts bbones to real bones and distributes the weights to the mesh"""
    bl_idname = "convert.bbones"
    bl_label = "convert bbones to bones"
    mesh: bpy.props.StringProperty()
    segmentPrefix: bpy.props.StringProperty(default= "")
    segmentSufix: bpy.props.StringProperty(default= ".segment")

    @classmethod
    def poll(cls,context):
        if context.mode == 'EDIT_ARMATURE':
            return context.selected_bones 

    def execute(self, context):
        selection = []
        
        for bone in context.selected_bones:
            if bone.bbone_segments != 1:
                selection.append(bone.name)
            
        for boneName in selection:
            bone = context.object.data.edit_bones[boneName]
            bboneSegments = subdivBone(context.object, bone, bone.bbone_segments - 1, prefix = self.segmentPrefix, sufix = self.segmentSufix)
            
            if context.scene.objects[self.mesh].vertex_groups.get(bone.name):
                distribute_vertWeights(context.object, context.scene.objects[self.mesh], bboneSegments, bone.name)
                
            else:
                print("R.U.N: no weight group found")
                
            for i, segment in enumerate(bboneSegments):    
                follow_bbone(context.object, segment, boneName,f = i/(len(bboneSegments)))
 

        return {'FINISHED'}

    def invoke(self, context, event):
        wm = context.window_manager
        return wm.invoke_props_dialog(self, width=200)

    def draw(self, context):

        layout = self.layout
        layout.label(text="convert bbones to bones")
        layout = self.layout.row()
        layout.label(text="target mesh")
        layout.prop_search(self, "mesh", bpy.data, "objects", text="")
        layout = self.layout.row()
        layout.label(text="segment prefix")
        layout.prop(self, 'segmentPrefix', text="")
        layout = self.layout.row()
        layout.label(text="segmanet sufix")
        layout.prop(self, 'segmentSufix', text="")
        layout = self.layout.row()


class VIEW3D_main_UI(bpy.types.Panel):
    bl_space_type = 'VIEW_3D'
    bl_region_type = 'UI'
    bl_category = "bbonesTool"
    bl_label = "Bbone tools"
    bl_idname = "MAIN_PT_UI"
    bl_order = 0

    @classmethod
    def poll(cls,context):

        if context.object and context.mode == 'EDIT_ARMATURE':
            return True 

    def draw(self, context):
        layout = self.layout
        layout = layout.column()
        layout.prop(context.active_bone, 'bbone_segments', text="segments")
        layout.label(text="Bbones to bones converter")
        layout.operator('convert.bbones', icon= 'CON_SPLINEIK')
        layout.label(text="set bbone handles")
        if context.active_bone.bbone_segments > 1:
            layout.prop(context.active_bone, 'bbone_handle_type_start', text="start")
             
            
            if context.selected_bones and len(context.selected_bones) == 2:
                   layout = self.layout.column()
                   layout.operator('set.bbonehandle0', icon= 'CON_SPLINEIK', text = " set selected as handle")
                   
            layout.prop(context.active_bone, 'bbone_handle_type_end', text="end")
            if context.selected_bones and len(context.selected_bones) == 2:
                layout.operator('set.bbonehandle1', icon= 'CON_SPLINEIK', text = " set selected as handle" )
                
        
               
        
        pass

def drawcontextMenu(self, context):
        layout = self.layout
        self.layout.separator()
        layout.prop(context.active_bone, 'bbone_segments', text="segments")
        if context.active_bone.bbone_segments > 1:
            layout.label(text = "Bbone start handle")
            layout.prop(context.active_bone, 'bbone_handle_type_start', text="")
            if context.selected_bones and len(context.selected_bones) == 2:
                   layout = self.layout.column()
                   layout.operator('set.bbonehandle0', icon= 'CON_SPLINEIK', text = " set start handle")
                   
            layout.label(text = "end handle")       
            layout.prop(context.active_bone, 'bbone_handle_type_end', text="")
            if context.selected_bones and len(context.selected_bones) == 2:
                layout.operator('set.bbonehandle1', icon= 'CON_SPLINEIK', text = " set end handle" )
        

blenderClasses = [
convert_bbonesOperator,
set_bboneEndHandleOperator,
set_bboneStartHandleOperator,
VIEW3D_main_UI

]

def register():
    for blender_class in blenderClasses:
        bpy.utils.register_class(blender_class)

    bpy.types.VIEW3D_MT_armature_context_menu.append(drawcontextMenu) 
        
        
def unregister():
    for blender_class in blenderClasses:
        bpy.utils.unregister_class(blender_class)
    bpy.types.VIEW3D_MT_armature_context_menu.remove(drawcontextMenu)
    
    print("hello world")


if __name__ == "__main__":
        register()